@@ -1,4 +1,6 @@
 from perceiver.data.text.bookcorpus import BookCorpusDataModule
+from perceiver.data.text.bookcorpusopen import BookCorpusOpenDataModule
+from perceiver.data.text.c4 import C4DataModule
 from perceiver.data.text.common import Task, TextPreprocessor
 from perceiver.data.text.enwik8 import Enwik8DataModule
 from perceiver.data.text.imdb import ImdbDataModule
