@@ -4,6 +4,7 @@ import jsonargparse
 
 from perceiver.data.text import (
     BookCorpusDataModule,
+    BookCorpusOpenDataModule,
     Enwik8DataModule,
     ImdbDataModule,
     Task,
@@ -14,6 +15,7 @@ from perceiver.data.text import (
 
 DATAMODULE_CLASSES = {
     "bookcorpus": BookCorpusDataModule,
+    "bookcorpusopen": BookCorpusOpenDataModule,
     "enwik8": Enwik8DataModule,
     "imdb": ImdbDataModule,
     "wikipedia": WikipediaDataModule,
@@ -39,5 +41,6 @@ if __name__ == "__main__":
     parser.add_argument("--mask_words", default=True, type=bool)
     parser.add_argument("--static_masking", default=False, type=bool)
     parser.add_argument("--add_special_tokens", default=False, type=bool)
+    parser.add_argument("--random_train_shift", default=False, type=bool)
     parser.add_argument("--preproc_workers", default=mp.cpu_count(), type=int)
     main(parser.parse_args())
